/* Copyright (c)  2003-2005 QLogic Corporation
*  QLogic Linux Fibre Channel Driver
*
*This program includes a device driver for Linux - 2.4.x that may be 
*distributed with QLogic hardware specific firmware binary file.  You may modify
*and redistribute the device driver code under the GNU General Public License as
*published by the Free Software Foundation (version 2 or a later version) and/or
*under the following terms, as applicable:
*
*	1. Redistribution of source code must retain the above copyright
*	   notice, this list of conditions and the following disclaimer.
*
*	2. Redistribution in binary form must reproduce the above copyright 
*	   notice, this list of conditions and the following disclaimer in the 
*	   documentation and/or other materials provided with the distribution.
*
*	3. The name of QLogic Corporation may not be used to endorse or promote 
*	   products derived from this software without specific prior written 
*	   permission.
*	
*You may redistribute the hardware specific firmware binary file under the 
*following terms:
*	1. Redistribution of source code (only if applicable), must retain the 
*	   above copyright notice, this list of conditions and the following 
*	   disclaimer.
*	2. Redistribution in binary form must reproduce the above copyright 
*	   notice, this list of conditions and the following disclaimer in 
*	   the documentation and/or other materials provided with the 
*	   distribution.
*	3. The name of QLogic Corporation may not be used to endorse or promote
*	   products derived from this software without specific prior written 
*	   permission.
*
*	REGARDLESS OF WHAT LICENSING MECHANISM IS USED OR APPLICABLE, THIS 
*	PROGRAM IS PROVIDED BY QLOGIC CORPORATION "AS IS'' AND ANY EXPRESS 
*	OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
*	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
*	ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, 
*	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
*	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
*	 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
*	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
*	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
*	IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*	POSSIBILITY OF SUCH DAMAGE.
*
*	USER ACKNOWLEDGES AND AGREES THAT USE OF THIS PROGRAM WILL NOT CREATE 
*	OR GIVE GROUNDS FOR A LICENSE BY IMPLICATION, ESTOPPEL, OR OTHERWISE 
*	IN ANY INTELLECTUAL PROPERTY RIGHTS (PATENT, COPYRIGHT, TRADE SECRET, 
*        MASK WORK, OR OTHER PROPRIETARY RIGHT) EMBODIED IN ANY OTHER QLOGIC 
*	HARDWARE OR SOFTWARE EITHER SOLELY OR IN COMBINATION WITH THIS PROGRAM.
*/


/****************************************************************************
              Please see revision.notes for revision history.
*****************************************************************************/

#if !defined(_QLA2XIP_H_)
#define _QLA2XIP_H_

#define WWN_SIZE		8
#define MAX_ADAPTER_COUNT	8	/* Maximum adapters supported */

#define MAX_MTU_SIZE		65280	/* Maximum MTU size */
#define MIN_MTU_SIZE		100	/* Minimum MTU size */
#define DEFAULT_MTU_SIZE	4096	/* Default MTU size */
#define DEFAULT_BUFFER_SIZE	(DEFAULT_MTU_SIZE + sizeof(PACKET_HEADER))
#define DEFAULT_RECEIVE_BUFFERS	32	/* Default number of receive buffers */

#define RECEIVE_BUFFERS_LOW_MARK 16	/* Receive buffers low water mark */
#define RECEIVE_BUFFERS_ADD_MARK 10	/* Receive buffers add mark */
#define DEFAULT_HEADER_SPLIT	0	/* Default header split size (== 0) */

#define LSD(x)	((uint32_t)((uint64_t)(x)))
#define MSD(x)	((uint32_t)((((uint64_t)(x)) >> 16) >> 16))

/*
 * Struct private for the QLogic IP adapter
 */
struct qla2xip_private {
	/* Net device members */
	struct net_device *next;
	spinlock_t lock;

	struct net_device_stats stats;	/* Device statistics */
	struct net_device *dev;	/* Parent NET device */

	uint32_t mtu;		/* Maximum transfer unit */
	uint16_t header_size;	/* Split header size */

	/* Send control block queue */
	struct send_cb *send_q[MAX_SEND_PACKETS + 1];
	uint16_t send_q_in;	/*  free in-pointer */
	uint16_t send_q_out;	/*  free out-pointer */
	/* Send control block array */
	struct send_cb send_buffers[MAX_SEND_PACKETS];
	struct packet_header *scb_header;
	dma_addr_t scb_header_dma;

	/* Inquiry data from SCSI driver */
	unsigned long options;	/* QLA2X00 supported options */
	void *ha;		/* Driver ha pointer */
	uint16_t link_speed;	/* Link speed */
	uint8_t port_name[WWN_SIZE];	/* Adapter port name */
	struct pci_dev *pdev;	/* PCI device information */

	/* Pointers to SCSI-backdoor callbacks */
	int (*ip_enable_routine) (void *, struct bd_enable *);
	int (*ip_disable_routine) (void *);
	int (*ip_add_buffers_routine) (void *, uint16_t, int);
	int (*ip_send_packet_routine) (void *, struct send_cb *);
	int (*ip_tx_timeout_routine) (void *);

	/* RISC receive queue */
	uint16_t receive_q_size;	/*  size */
	struct risc_rec_entry *receive_q;	/*  pointer */
	struct risc_rec_entry *receive_q_in;	/*  in-pointer */
	struct risc_rec_entry *receive_q_end;	/*  end-pointer */
	uint16_t receive_q_cnt;	/*  current buffer count */
	uint16_t receive_q_add_cnt;	/*  buffers to be added */

	struct buffer_cb receive_buffers[MAX_RECEIVE_BUFFERS];
	uint16_t max_receive_buffers;	/*  maximum # receive buffers */
	uint32_t receive_buff_data_size;	/*  data size */
};
#endif
